@@ -2,13 +2,13 @@
 
 import copy
 import ete3
-import numpy as np
 import networkx as nx
+import newick
+import numpy as np
 import warnings
 from cvxopt.solvers import qp as cvxopt_qp
 from cvxopt import matrix as cvxopt_matrix
 from numbers import Number
-import sys
 
 import lineageot.simulation as sim
 
@@ -258,7 +258,7 @@ def add_times(tree, mutation_rates, time_inference_method, overwrite = False):
         # root should have time zero
         assert(tree.nodes[leaves[-1]]['time'] == 0)
 
-        laplacian = nx.linalg.laplacian_matrix(tree.to_undirected(),
+        laplacian = nx.linalg.laplacian_matrix(nx.Graph(tree),
                                                nodelist = internal_nodes + leaves)
         incidence_matrix = nx.linalg.incidence_matrix(tree,
                                                       nodelist = internal_nodes + leaves,
@@ -387,15 +387,17 @@ def add_node_times_from_dict(tree, current_node, time_dict):
 
     return
 
-def add_node_times_from_division_times(tree, current_node = 'root'):
+def add_node_times_from_division_times(tree, current_node = 'root', overwrite = False):
     """
     Adds 'time' variable to all descendants of current_node based on the 'time_to_parent' variable
     """
     if current_node == 'root':
-        tree.nodes[current_node]['time'] = 0
+        if overwrite | (not 'time' in tree.nodes[current_node]):
+            tree.nodes[current_node]['time'] = 0
     else:
         parent = next(tree.predecessors(current_node))
-        tree.nodes[current_node]['time'] = tree.nodes[parent]['time'] + tree.nodes[current_node]['time_to_parent']
+        if overwrite | (not 'time' in tree.nodes[current_node]):
+            tree.nodes[current_node]['time'] = tree.nodes[parent]['time'] + tree.nodes[current_node]['time_to_parent']
 
     children = tree.successors(current_node)
     for child in children:
@@ -674,7 +676,7 @@ def get_lineage_distances_across_time(early_tree, late_tree):
 
     # get length of path up to parent of early_cell and back to early_cell
     for late_cell in cells_late:
-        distance_dictionary, tmp = nx.single_source_dijkstra(late_tree.to_undirected(),
+        distance_dictionary, tmp = nx.single_source_dijkstra(nx.Graph(late_tree),
                                                             late_cell,
                                                             weight = 'time')
         for early_cell in cells_early:
@@ -771,7 +773,29 @@ def split_edge(tree, edge, new_node):
 
 
 
+def get_components(graph, edge_length_key = 'time'):
+    """
+    Returns subgraph views corresponding to connected components of the graph
+    if edges of infinite length are removed
+    
+    Parameters
+    ----------
+    graph: NetworkX graph
+    edge_length_key: default 'time'
 
+    Returns
+    -------
+    subgraphs: List of NetworkX subgraph views
+    """
+    # copying graph as undirected
+    g = nx.Graph(graph)
+    
+    edges_to_remove = [e for e in g.edges if g.edges[e][edge_length_key] == np.inf]
+    g.remove_edges_from(edges_to_remove)
+    
+    component_node_sets = nx.connected_components(g)
+
+    return [graph.subgraph(component_nodes) for component_nodes in component_node_sets] 
 
 
 
@@ -790,15 +814,24 @@ def add_conditional_means_and_variances(tree, observed_nodes):
     nodes, conditional on the observed values of 'x' in observed_nodes,
     assuming that differences along edges are Gaussian with variance equal to 
     the length of the edge.
+
+    In doing so, also adds inverse time annotations to edges.
+
+    If no nodes in tree are observed, inverse time annotations are added but
+    conditional means and variances are not (as there is nothing to condition on).
     """
     node_list = [n for n in tree.nodes]
     
     add_inverse_times_to_edges(tree)
-    l = nx.laplacian_matrix(tree.to_undirected(), nodelist = node_list, weight = 'inverse time')
+    l = nx.laplacian_matrix(nx.Graph(tree), nodelist = node_list, weight = 'inverse time')
     
     unobserved_nodes = [n for n in node_list if not n in observed_nodes]
     # Resorting so the order of indices in all matrices match
+    # and removing observed_nodes not in tree
     observed_nodes = [n for n in node_list if n in observed_nodes]
+    if len(observed_nodes) == 0:
+        # no observed nodes in this component
+        return
     unobserved_node_indices = [i for i in range(len(node_list)) if not node_list[i] in observed_nodes]
     observed_node_indices = [i for i in range(len(node_list)) if node_list[i] in observed_nodes]
     
@@ -846,12 +879,219 @@ def get_ancestor_data(tree, time, leaf = None):
 
 
 
+##########################
+# Tree fitting functions #
+##########################
+
+def make_tree_from_nonnested_clones(clone_matrix, time, root_time_factor = 1000):
+    """
+    Creates a forest of stars from clonally-labeled data. The centers of the stars are connected to a root far in the past.
+
+    Parameters
+    ----------
+    clone_matrix: Boolean array with shape [num_cells, num_clones]
+        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
+        Each cell should belong to exactly one clone.
+    time: Number
+        The time of sampling of cells relative to initial clonal labelling.
+    root_time_factor: Number, default 1000
+        Relative time to root of tree (i.e., most recent common ancestor of all cells).
+        The time of the root is set to -root_time_factor*time.
+        The default is large so minimal information is shared across clones.
+    Returns
+    -------
+    fitted_tree: NetworkX DiGraph
+        A tree annotated with edge and node times
+    """
+    fitted_tree = nx.DiGraph()
+
+    n_cells, n_clones = clone_matrix.shape
+    clone_labels = [-i-1 for i in range(n_clones) if np.sum(clone_matrix[:, i]) > 0] # negative so distinct from cells, skipping empty clones
+    
+    # add nodes for cells
+    fitted_tree.add_nodes_from(range(n_cells), time = time, time_to_parent = time)
+
+    # add nodes for clone progenitors
+    fitted_tree.add_nodes_from(clone_labels, time = 0, time_to_parent = root_time_factor*time)
+
+    # add edges from clone progenitors to observed cells
+    for clone in clone_labels:
+        cells_in_clone = np.nonzero(clone_matrix[:, -clone - 1])[0]
+        fitted_tree.add_edges_from([(clone, cell, {'time': time}) for cell in cells_in_clone])
+
+    # add root
+    fitted_tree.add_node('root', time = -root_time_factor*time)
+
+    # add edges to root
+    fitted_tree.add_edges_from([('root', i, {'time': root_time_factor*time}) for i in clone_labels])
+
+    return fitted_tree
 
 
+def make_tree_from_clones(clone_matrix, time, clone_times, root_time = -np.inf):
+    """
+    Adds a leaf for each row in clone_matrix to clone_reference_tree. The parent is set as the
+    clone that the cell is a member of with the latest labeling time.
+
+    clone_reference_tree is edited in place rather than returned.
+
+    Parameters
+    ----------
+    clone_matrix: Boolean array with shape [num_cells, num_clones]
+        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
+    clone_times: Vector of length num_clones
+        Each entry has the time of labeling of the corresponding clone.
+    time: Number
+        The time of sampling of cells.
+    root_time: Number, default -np.inf
+        The time of the most recent common ancestor of all clones. 
+        If -np.inf, clones are effectively treated as unrelated
+
+    Returns
+    -------
+    fitted_tree: NetworkX DiGraph
+        A tree annotated with edge and node times
+    """
+    fitted_tree = make_clone_reference_tree(clone_matrix, clone_times, root_time)
+    add_samples_to_clone_tree(clone_matrix, clone_times, fitted_tree, time)
+    return fitted_tree
 
 
+def add_samples_to_clone_tree(clone_matrix, clone_times, clone_reference_tree, sampling_time):
+    """
+    Adds a leaf for each row in clone_matrix to clone_reference_tree. The parent is set as the
+    clone that the cell is a member of with the latest labeling time.
+
+    clone_reference_tree is edited in place rather than returned.
+
+    Parameters
+    ----------
+    clone_matrix: Boolean array with shape [num_cells, num_clones]
+        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
+    clone_times: Vector of length num_clones
+        Each entry has the time of labeling of the corresponding clone.
+    clone_reference_tree: 
+        The tree of lineage relationships among clones.
+    sampling_time: Number
+        The time of sampling of the cells. Should be greater than all clone labeling times.
+
+    Returns
+    -------
+    """
+    n_cells, n_clones = clone_matrix.shape
+    for cell in range(n_cells):
+        parent_index = get_parent_clone_of_leaf(cell, clone_matrix, clone_times)
+        parent_label = 'clone_' + str(parent_index)
+        edge_time = sampling_time - clone_reference_tree.nodes[parent_label]['time']
+        clone_reference_tree.add_node(cell, time = sampling_time, time_to_parent = edge_time)
+        clone_reference_tree.add_edge(parent_label, cell, time = edge_time)
+    return
 
 
+def get_parent_clone_of_leaf(leaf, clone_matrix, clone_times):
+    """
+    Returns the index of the clone that the leaf is a member of with the latest labeling time.
+    """
+    candidate_clones = np.nonzero(clone_matrix[leaf, :])[0]
+    candidate_times = clone_times[candidate_clones]
+    maximal_time_indices = np.nonzero(candidate_times == np.max(candidate_times))[0]
+
+    if len(maximal_time_indices) != 1:
+        raise ValueError("Could not unambiguously assign a sampled cell to a parent clone.")
+    else:
+        return candidate_clones[maximal_time_indices[0]]
+
+
+def make_clone_reference_tree(clone_matrix, clone_times, root_time = -np.inf):
+    """
+    Makes a tree with nodes for each clone.
+    
+    Parameters
+    ----------
+    clone_matrix: Boolean array with shape [num_cells, num_clones]
+        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
+    clone_times: Vector of length num_clones
+        Each entry has the time of labeling of the corresponding clone.
+    root_time: Number, default -np.inf
+        The time of the most recent common ancestor of all clones. 
+        If -np.inf, clone subtrees are effectively treated independently.
+
+    Returns
+    -------
+    clone_reference_tree: NetworkX DiGraph
+        A tree of clones (not sampled cells), annotated with edge and node times
+    """
+
+    n_cells, n_clones = clone_matrix.shape
+    unique_times = np.unique(clone_times) # output is sorted
+    clone_labels = ['clone_' + str(i) for i in range(n_clones)]
+
+    # make base level of tree
+    clone_reference_tree = nx.DiGraph()
+    clone_reference_tree.add_node('root', time = root_time)
+
+    current_clone_indices = np.nonzero(clone_times == unique_times[0])[0]
+    current_clone_labels = [clone_labels[i] for i in current_clone_indices]
+    clone_reference_tree.add_nodes_from(current_clone_labels, time = unique_times[0], time_to_parent = unique_times[0] - root_time)
+    clone_reference_tree.add_edges_from([('root', label, {'time': unique_times[0] - root_time}) for label in current_clone_labels])
+
+    # make subsequent levels
+    for level, labeling_time in zip(range(1, len(unique_times)), unique_times[1:]):
+        current_clone_indices = np.nonzero(clone_times == labeling_time)[0]
+        current_clone_labels = [clone_labels[i] for i in current_clone_indices]
+
+        clone_reference_tree.add_nodes_from(current_clone_labels, time = labeling_time)
+
+        for clone_index in current_clone_indices:
+            parent_index = find_parent_clone(clone_index, clone_matrix, clone_times)
+            if parent_index == 'root':
+                parent_label = 'root'
+            else:
+                parent_label = clone_labels[parent_index]
+
+            time_to_parent = labeling_time - clone_reference_tree.nodes[parent_label]['time']
+            
+            clone_reference_tree.nodes[clone_labels[clone_index]]['time_to_parent'] = time_to_parent
+            clone_reference_tree.add_edge(parent_label, clone_labels[clone_index], time = time_to_parent)
+
+    return clone_reference_tree
+
+
+def find_parent_clone(clone, clone_matrix, clone_times):
+    """
+    Returns the parent of a subclone, assuming this is uniquely defined as the clone from
+    an earlier time point whose barcode was observed in a cell from the subclone.
+
+    Parameters
+    ----------
+    clone: int
+        Index of clone whose parent will be returned
+    clone_matrix: Boolean array with shape [num_cells, num_clones]
+        Each entry is 1 if the corresponding cell belongs to the corresponding clone and zero otherwise.
+    clone_times: Vector of length num_clones
+        Each entry has the time of labeling of the corresponding clone.
+
+    Returns
+    -------
+    parent: int
+       Index of parent clone.
+    """
+    clone_is_earlier = clone_times < clone_times[clone]
+    clone_is_shared = np.array(np.sum(clone_matrix[np.nonzero(clone_matrix[:, clone] == 1)[0], :], 0) > 0)
+    clone_is_shared = np.reshape(clone_is_shared, clone_is_shared.size) # does nothing if everyhing was ndarray, but corrects an issue with sparse matrices
+
+    candidates = np.nonzero(clone_is_earlier & clone_is_shared)[0]
+
+    if len(candidates) == 1:
+        return candidates[0]
+    elif len(candidates) == 0:
+        # no parent clone observed
+        return 'root'
+    else:
+        print(clone_is_earlier)
+        print(clone_is_shared)
+        print(clone, candidates)
+        raise ValueError("The input clone matrix does not unambiguously define a clonal lineage tree. Please use another method to create the tree.")
 
 
 ###################################
@@ -877,81 +1117,69 @@ def neighbor_join(distance_matrix):
     """
     n = distance_matrix.shape[0]
 
-    leaf_nodes = []
+    assert(n >= 3) # no need to bother with neighbor joining if one or two nodes
+
+    unjoined_nodes = []
     for i in range(n-1):
         subtree = nx.DiGraph()
         subtree.add_node(i)
         node = NeighborJoinNode(subtree, i, False)
-        leaf_nodes.append(node)
+        unjoined_nodes.append(node)
 
     subtree = nx.DiGraph()
     subtree.add_node('root')
     node = NeighborJoinNode(subtree, 'root', True)
-    leaf_nodes.append(node)
+    unjoined_nodes.append(node)
 
-    initial_recursion_limit = sys.getrecursionlimit()
-    if n > initial_recursion_limit:
-        warnings.warn("Temporarily increasing recursion limit for neighbor joining.")
-        sys.setrecursionlimit(n + 50) # number of nodes plus an arbitary buffer
+    next_node_to_add = -1 # integer label of next node to add, will count down
+    # (leaf nodes have positive integer labels, root is labeled 'root')
 
-    fitted_tree = recursive_neighbor_join(distance_matrix, leaf_nodes, -1)
+    while len(unjoined_nodes) > 3:
+        # recursive part of neighbor joining algorithm
 
-    if n > initial_recursion_limit:
-        sys.setrecursionlimit(initial_recursion_limit)
+        # Compute Q matrix
+        Q = compute_q_matrix(distance_matrix)
 
-    return fitted_tree
+        # Pick nodes to join
+        nodes_to_join = pick_joined_nodes(Q)
+        
+        distances = distances_to_joined_node(distance_matrix, nodes_to_join)
 
-
-def recursive_neighbor_join(distance_matrix, nodes, next_node_to_add):
-    """
-    Recursive helper function for neighbor joining
-
-    distance_matrix:  array of pairwise distances between nodes
-    nodes:            list of NeighborJoinNode nodes to be joined
-    next_node_to_add: integer label of next node to add (negative)
-    """
-
-    # Base case: if there are three nodes, join them all
-    # (and return only the tree)
-    if len(nodes) == 3:
-        distances = distances_to_joined_node(distance_matrix, [0,1])
-        last_join = join_nodes(nodes[0], nodes[1], next_node_to_add, distances)
+        # Remove those nodes and add merged node
+        new_node = join_nodes(unjoined_nodes[nodes_to_join[0]], unjoined_nodes[nodes_to_join[1]], next_node_to_add, distances)
         next_node_to_add = next_node_to_add - 1
         
-        last_edge_distance = (distance_matrix[2, 0] 
-                              + distance_matrix[2, 1]
-                              - distance_matrix[1, 0])/2
-        T = nx.compose(nodes[2].subtree, last_join.subtree)
+        unjoined_nodes = [unjoined_nodes[i] for i in range(len(unjoined_nodes)) if i not in nodes_to_join]
+        unjoined_nodes.append(new_node)
 
-        if last_join.has_global_root:
-            assert(not nodes[2].has_global_root)
-            T.add_edge(last_join.subtree_root, nodes[2].subtree_root)
-            T.nodes[nodes[2].subtree_root]['time_to_parent'] = last_edge_distance
-        else:
-            assert(nodes[2].has_global_root)
-            T.add_edge(nodes[2].subtree_root, last_join.subtree_root)
-            T.nodes[last_join.subtree_root]['time_to_parent'] = last_edge_distance
-        return T
+        # Compute new distances
+        distance_matrix = compute_new_distances(distance_matrix, nodes_to_join)
 
-    # Compute Q matrix
-    Q = compute_q_matrix(distance_matrix)
 
-    # Pick nodes to join
-    nodes_to_join = pick_joined_nodes(Q)
 
-    distances = distances_to_joined_node(distance_matrix, nodes_to_join)
-
-    # Remove those nodes and add merged node
-    new_node = join_nodes(nodes[nodes_to_join[0]], nodes[nodes_to_join[1]], next_node_to_add, distances)
+    # only three remaining nodes now; join an arbitrary two of them directly
+    distances = distances_to_joined_node(distance_matrix, [0,1])
+    last_join = join_nodes(unjoined_nodes[0], unjoined_nodes[1], next_node_to_add, distances)
     next_node_to_add = next_node_to_add - 1
+    
+    last_edge_distance = (distance_matrix[2, 0] 
+                          + distance_matrix[2, 1]
+                          - distance_matrix[1, 0])/2
+    fitted_tree = nx.compose(unjoined_nodes[2].subtree, last_join.subtree)
+    
+    # add 'time_to_parent' to the node that is not the root
+    if last_join.has_global_root:
+        assert(not unjoined_nodes[2].has_global_root)
+        fitted_tree.add_edge(last_join.subtree_root, unjoined_nodes[2].subtree_root)
+        fitted_tree.nodes[unjoined_nodes[2].subtree_root]['time_to_parent'] = last_edge_distance
+    else:
+        assert(unjoined_nodes[2].has_global_root)
+        fitted_tree.add_edge(unjoined_nodes[2].subtree_root, last_join.subtree_root)
+        fitted_tree.nodes[last_join.subtree_root]['time_to_parent'] = last_edge_distance
 
-    new_nodes = [nodes[i] for i in range(len(nodes)) if i not in nodes_to_join]
-    new_nodes.append(new_node)
+    
+    return fitted_tree
 
-    # Compute new distances
-    new_distances = compute_new_distances(distance_matrix, nodes_to_join)
-
-    return recursive_neighbor_join(new_distances, new_nodes, next_node_to_add)
 
         
 
@@ -1085,10 +1313,87 @@ def robinson_foulds(tree1, tree2):
 
 
 
+####################################
+# Converting trees between formats #
+####################################
 
+def convert_newick_to_networkx(newick_tree, leaf_labels, leaf_time = None, root_label = 'root', unlabeled_nodes_added = 0, at_global_root = True):
+    """
+    Converts a tree from the Newick package's format to LineageOT's annotated NetworkX DiGraph. Ignores existing annotations, except edge lengths.
 
+    Parameters
+    ----------
+    newick_tree : newick.Node or [newick.Node]
+        A tree loaded by the Newick package.
+    leaf_labels : list
+        The label of each leaf in the Newick tree, sorted to align with the gene expression AnnData object filtered to cells at the corresponding time
+    leaf_time : float (default None)
+        The time of sampling of the leaves. If unspecified, the root of the tree is assigned time 0.
+    root_label : str (default 'root')
+        The label of the root node of the tree
+    unlabeled_nodes_added : int (default 0)
+        The number of previously-unlabeled nodes that have already been added to the tree. Leave as 0 for any top-level use of the function.
+    at_global_root : bool (default True)
+        Whether the function is being called to convert a full tree or a subtree.
+    Returns
+    -------
+    tree : Networkx DiGraph
+        The saved tree, in LineageOT's format.
+        Each node is annotated with 'time_to_parent' and 'time' (which indicates either the time of sampling (for observed cells) or the time of division (for unobserved ancestors)).
+        Edges are directed from parent to child and are annotated with 'time' equal to the child node's 'time_to_parent'.
+        Observed node indices correspond to their index in leaf_labels.
+    """
+    assert(type(newick_tree) == newick.Node)
+    tree = nx.DiGraph()
+    if root_label == None:
+        # node was unlabeled in original Newick tree
+        # give it a new unique name
+        root_label = "Unlabeled_" + str(unlabeled_nodes_added)
+        unlabeled_nodes_added = unlabeled_nodes_added + 1
+    root_name = root_label
+    if len(newick_tree.descendants) == 0:
+        # at a leaf, use index of label as name
+        if not (root_label in leaf_labels):
+            raise ValueError("One of the leaves in the Newick tree does not have a corresponding label in leaf_labels")
+        root_label = leaf_labels.index(root_label)
+        
+    tree.add_node(root_label, name = root_name)
+    
+    for child in newick_tree.descendants:
+        # convert tree for child
+        subtree, unlabeled_nodes_added, new_child_label = convert_newick_to_networkx(child, leaf_labels, leaf_time = leaf_time, root_label = child.name, unlabeled_nodes_added = unlabeled_nodes_added, at_global_root = False)
+        
+        # add to existing tree
+        tree = nx.union(tree, subtree) # not sure this is efficient
+        
+        # annotate edge times
+        tree.add_edge(root_label, new_child_label, time=child.length)
+        tree.nodes[new_child_label]['time_to_parent'] = child.length
 
+    # annotate node times
 
+    if leaf_time == None:
+        # reference time is global root, do nothing in recursive case
+        if at_global_root:
+            tree.nodes[root_label]['time'] = 0
+            add_node_times_from_division_times(tree, current_node = root_label) 
+    else:
+        if len(newick_tree.descendants) == 0:
+            # at leaf
+            tree.nodes[root_label]['time'] = leaf_time
+        else:
+            # add time annotations backwards
+            times_from_children = np.array([tree.nodes[child]['time'] - tree.nodes[child]['time_to_parent'] for child in tree.successors(root_label)])
+            if np.all(times_from_children == times_from_children[0]):
+                # time labels from different branches are consistent
+                tree.nodes[root_label]['time'] = times_from_children[0]
+            else:
+                raise ValueError("Edge length annotations in Newick tree are not consistent with a single sampling time for all leaves. Consider using the global root as a reference time.")
+
+    if at_global_root:
+        return tree
+    else:
+        return tree, unlabeled_nodes_added, root_label
 
 
 def tree_to_ete3(tree):
